package PureMessage::HTTPD::Services::Quarantine;
use strict;
use warnings;
use base qw( PureMessage::HTTPD::Services::Application );
use PureMessage::Version qw($PREFIX);
use PureMessage::_sys;
use PureMessage::Quarantine;
use PureMessage::HTTPD::Services::Access qw(ip_is_valid);
use PureMessage::Config qw(%config);
use ActiveState::MIME::Parser;
use Apache::Constants qw(FORBIDDEN);
use File::Path;
use File::Basename;
use Storable qw (nfreeze);

my $CHUNK_SIZE = 10240;

# Setup; specify name of logfile to use and list of available run modes.
sub setup {
    my $self = shift;
    $self->set_logfile( 'service_quarantine_log' );
    $self->start_mode('message');
    $self->run_modes(
        # Quarantine runmodes
        message     => 'show_message',
        attach_info => 'show_file_attachments',
        attachment  => 'send_file_attachment',
        source      => 'show_source',
        minfo       => 'show_minfo',

        # MTA Info runmodes
        qshape      => 'show_qshape',

        # Other runmodes
        denied      => 'deny_service',
    );
    $self->header_props(-type => 'text/plain');
}

sub cgiapp_prerun {
    my $self = shift;

    my $r = $self->param('r') || die "Can't find apache request object!";
    my $c = $r->connection;
    my $ip = $c->remote_ip;
    unless (ip_is_valid($ip)) {
        $self->prerun_mode('denied');
    }
}

sub show_source {
    my ($self) = @_;
    my $global_id = $self->_find_global_id;
    $self->log("Request for source $global_id");
    my $msg = $self->_fetch_message_obj($global_id, 'dir');

    # XXX - If possible, we should probably make this streaming
    my $source = $msg->message;
    return $source;
}

sub show_message {
    my ($self) = @_;
    my $global_id = $self->_find_global_id;
    $self->log("Request for message $global_id");
    my $msg = $self->_fetch_message_obj($global_id, 'dir');
    return $msg->cleaned_body;
}

sub show_minfo {
    my ($self) = @_;
    my $global_id = $self->_find_global_id;
    $self->log("Request for message $global_id");
    my $msg = $self->_fetch_message_obj($global_id, 'dir');
    my @msg_minfo = $msg->minfo_items;
    return nfreeze \@msg_minfo;
}

sub _truncate_match_text {
    my $string = shift;
    my $suf = length($string) > 15 ? '...' : '';
    $string = substr($string,0,15) . $suf;
    return $string;
}

sub _truncate_filename {
    my $string = shift;
    my $pre = length($string) > 35 ? '...' : '';
    $string = $pre . substr($string,-35);
    return $string;
}

sub show_file_attachments {
    my ($self) = shift;
    my $global_id = $self->_find_global_id;
    $self->log("Request for attachment info $global_id");

    my $msg = $self->_fetch_message_obj($global_id, 'dir');

    my $minfo = $msg->minfo();
    my $scan_attach = { };

    my %hash = ( );
    for(my $i = 0; $i < $#$minfo; $i++) {
         if($minfo->[$i] eq 'Scanned-Attachment') {

            my($filetype, $filename, $matchtext) = split(/\|/, $minfo->[++$i],3);

            $filetype  =~ s/^\s*(.*)\s$/$1/;
            $filetype  =~ s/\\n//;
            $filename  =~ s/^\s*(.*)\s$/$1/;
            $filename  =~ s/\\n//;
            $matchtext =~ s/^\s*(.*)\s$/$1/;
            $matchtext =~ s/\\n//;

            my $basename = basename($filename);

            my %m_info = (
                'filetype'     => $filetype,
                'fullfilename' => $filename,
                'filename'     => _truncate_filename($filename),
                'text'         => _truncate_match_text($matchtext),
                'text_size'    => length($matchtext),
            );

	    my $file;
	    my @junk;
	    ($file, @junk) = split(/\//, $filename);
            $scan_attach->{$file} ||= [];
            push(@{$scan_attach->{$file}}, \%m_info);
        }
    }

    my $attach_info = {};

    my $mp = ActiveState::MIME::Parser->new;
    #$mp->descend_multipart(1);
    #$mp->descend_sub_message(1);
   
    $mp->handler(part_start => sub {
        my ($p, $hdrs, $ct, $ct_param, $te) = @_;
        my $partid = $p->part_id();
        $attach_info->{$partid}{ct} = $ct;
        $attach_info->{$partid}{size} = 0;
        
        # Override filename with CD when CT & CD both have it.
        $attach_info->{$partid}{filename} = $ct_param->{name} if $ct_param->{name};
        # We cannot deal with multi-line filename.
        if($hdrs =~ /\nContent-Disposition:\s+(.*?)filename\s*=\s*"?([^\r\n;\"]+)/is) {
            $attach_info->{$partid}{filename} = $2 if $1 !~ /\r?\n\S/s;
        }

        # Skip this part if there is no filename available.
        delete $attach_info->{$partid} unless
            $attach_info->{$partid}{filename};
    });

    $mp->handler(data_decoded => sub {
        my ($p, $data) = @_;
        my $partid = $p->part_id;
        return unless exists $attach_info->{$partid};
        $attach_info->{$partid}{size} += length $data;
    });
    
    my $buf;
    while($msg->read($buf, $CHUNK_SIZE)) {
        $mp->parse($buf);
    }
    $mp->eof;

    my $output = q{};
    foreach my $partid (sort { $a <=> $b } keys %$attach_info){
        my ($size, $filename, $content_type) = @{ $attach_info->{$partid} }{qw(size filename ct)};
	$output .= "$partid|$filename|$content_type|$size\n";
        foreach my $minfo (@{$scan_attach->{$filename}}) {
            $output .= "$partid|$filename|$content_type|$size|$minfo->{filename}|$minfo->{fullfilename}|$minfo->{filetype}|$minfo->{text_size}|$minfo->{text}\n";
        }
    }
    
    return $output;
}

sub send_file_attachment {
    my ($self) = @_;
    my ($global_id, $part) = $self->_find_global_id;
    die "A part_id is mandatory for this service" unless $part;
    $self->log("Request for attachment $global_id part $part");

    my $msg = $self->_fetch_message_obj($global_id, 'dir');
    my $attach_info = {};

    # Tell CGI-App not to send headers:
    $self->header_type('none');

    my $mp = ActiveState::MIME::Parser->new;
    #$mp->descend_multipart(1);
    #$mp->descend_sub_message(1);

    # Write the decoded attachment to a temp file
    my $tmpdir = _create_tmp_dir();
    my $tmp = "$tmpdir/attachment.$$";
    open(my $wfh, ">$tmp") or die "Can't open $tmp: $!";
    binmode $wfh;

    my ($filename, $content_type);
   
    $mp->handler(part_start => sub {
        my ($p, $hdrs, $ct, $ct_param, $te) = @_;
        my $partid = $p->part_id();
        return unless $partid == $part;
        $content_type = $ct;
        
        # Override filename with CD when CT & CD both have it.
        $filename = $ct_param->{name} if $ct_param->{name};
        # We cannot deal with multi-line filename.
        if($hdrs =~ /\nContent-Disposition:\s+(.*?)filename\s*=\s*"?([^\r\n;\"]+)/is) {
            $filename = $2 if $1 !~ /\r?\n\S/s;
        }
    });
    $filename ||= 'attachment';

    $mp->handler(data_decoded => sub {
        my ($p, $data) = @_;
        return unless $p->part_id == $part;
        print $wfh $data;
    });
    
    my $buf;
    while($msg->read($buf, $CHUNK_SIZE)) {
        $mp->parse($buf);
    }
    $mp->eof;

    close $wfh or die "Can't write $tmp: $!";

    my $size = (stat($tmp))[7];
    print $self->query->header(
        -type => $content_type,
        '-Content-Disposition' => "attachment; filename=$filename",
        '-Content-Length' => $size,
    );
    # read and send file
    open(my $rfh, $tmp) or die "Can't open $tmp: $!";
    while(read($rfh, my $buffer, $CHUNK_SIZE)) {
        # quit early if the connection closes (print fails)
        last unless print $buffer;
    }
    close $tmp;
    _remove_tmp_dir();
        
    return '';
}

sub show_qshape {
    my ($self) = @_;
    my $path_info = $self->_path_info;
    unless ($path_info =~ m#^/qshape(?:/(\w+))?#) {
        die "Cannot handle qshape request: $path_info";
    }

    my $queue = $1 || '';
    $self->log("Request for qshape for queue '$queue'");
    my $qshape = "$PREFIX/postfix/sbin/pmx-qshape";
    die "Can't find qshape: $qshape" unless -x $qshape;
    return qx($qshape $queue);
}

sub deny_service {
    my $self = shift;

    # Tell CGI-App not to send any headers, tell Apache to 
    # send FORBIDDEN response code
    $self->header_type('none');
    my $r = $self->param('r');
    $r->status(FORBIDDEN);
    $r->send_http_header;
    return '';
}

sub _fetch_message_obj {
    my ($self, $id, $type) = @_;
    my $qdir = "dir:$config{quarantine_dir}"
        if defined $type and $type eq 'dir';
    my $mstore;
    if (defined $qdir) {
        $mstore = PureMessage::Quarantine->new($qdir);
    }else{
        $mstore = PureMessage::Quarantine->new;
    }
    die "Failed to create a new message store" unless $mstore;
    my $message = $mstore->message($id);
    die "Couldn't fetch message $id" unless $message;
    return $message->get_local_message;
}

sub _find_global_id {
    my $self = shift;
    my $path_info = $self->_path_info;
    if ($path_info =~ m#/(\d+-\d+)(?:/([\d\.]+\d))?$#) {
        return wantarray ? ($1, $2) : $1;
    }
    die "Can't find a global id in '$path_info'!";
}

sub _path_info {
    my $self = shift;
    my $path_info = $self->query->path_info;
    $path_info =~ s#^/quarantine##;
    return $path_info;
}

sub _tmp_dir {
    return "$PREFIX/var/tmp/attachments/$$";
}

sub _remove_tmp_dir {
    my $tmpdir = _tmp_dir();
    rmtree($tmpdir) if -e $tmpdir;
    return $tmpdir;
}

sub _create_tmp_dir {
    my $tmpdir = _tmp_dir();
    _remove_tmp_dir();
    mkpath($tmpdir) unless -e $tmpdir;
    return $tmpdir;
}

1;

=head1 NAME

PureMessage::HTTPD::Services::Quarantine - HTTPD Service for PureMessage Quarantine

=head1 COPYRIGHT

Copyright (C) 2000-2006 Sophos Group.  All rights reserved.  Sophos and
PureMessage are trademarks of Sophos Plc and Sophos Group.

=cut
